#ifndef	_FAR_H_
#define	_FAR_H_

#include "far_types.h"
#include "far_err.h"

typedef unsigned char char_t;
typedef unsigned short fver_t;

/*
 * far_stat_t: this structure is a portable version of the stat struct
 *             it features 64bit support and has less fields (unuseful
 *             to libfar) than stat
 */
typedef struct _far_stat
{
	uint32_t	fst_uid,
			fst_gid;
	uint32_t	fst_mode;
	uint32_t	__pad;
	uint64_t	fst_atime;
	uint64_t	fst_mtime;
	uint64_t	fst_rdev;
	uint64_t	fst_size;
} far_stat_t;

/*
 * far_list_t: this structure (used as a node in a linklist) contains all
 *             the information about an entry. In far v3  a similar
 *             structure was used as node in a linklist and to store
 *             the entry's description in the archive, because of that
 *             we had two 1024bytes fields to store path and linkpath
 *             and 2 unuseful pointers (next and prev), this
 *             could cause problems on an architecture with pointer's size
 *             different than 4bytes and resulted in big archives
 *             (the structure was 2158bytes, multiply this number for the
 *             number of entries in an archive and you'll find
 *             that's too much big), so I decided to use a big structure
 *             like this where store all the information to build an
 *             archive and a littler one to be used as part of the archive's
 *             header.
 */
typedef struct _far_info_list
{
	struct	_far_info_list	*next;		/* next entry */
	struct	_far_info_list	*prev;		/* previous entry */
	struct	_far_info_list	*sub;		/* directory's content */
	struct	_far_info_list	*parent;	/* upper directory */
	uint32_t		_pad;
	far_stat_t		*d_info;	/* stat info */
	char			*d_path;	/* entry's name */
	char			*d_lpath;	/* symlink's path */
	uint32_t		d_pathsize;	/* path size */
	uint32_t		d_lpathsize;	/* linkpath size */
	uint64_t		d_offset;
	char_t			d_cksum[16];	/* file's chksum */
	char_t			d_hash[16];	/* hash for faster searches */
} far_list_t, far_info_t;

/*
 * far_head_t: this structure describes an entry inside the archive,
 *             it's used as part of the archive's header
 */
typedef struct _far_header
{
	far_stat_t	d_info;		/* stat info about the entry */
	uint64_t	d_offset;	/* where the file start in the arch */
	uint32_t	d_pathsize;	/* path size */
	uint32_t	d_lpathsize;	/* linkpath size (if it's a link)*/
	char_t		d_hash[16];	/* hash for faster searches */
	char_t		d_cksum[16];	/* checksum of the file */
} far_head_t;

ferr_t far_buffer_archive(const char **, int, const char *, void **,
			uint64_t *);
ferr_t far_filedesc_archive(const char **, int, const char *, int);
#define	far_sockdesc_archive(x, y, z) far_filedesc_archive(x, y, z)

ferr_t far_buffer_extract(char_t *, uint64_t, const char *);
ferr_t far_filedesc_extract(int, const char *);
ferr_t far_sockdesc_extract(int);

#endif	/* _FAR_H_ */
