#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <regex.h>
#include <dirent.h>
#if (defined(__linux__) || defined(_linux_))
#	include <netinet/in.h>
#endif

#include "far_err.h"
#include "far_types.h"
#include "far.h"
#include "streams.h"
#include "archive.h"
#include "bits.h"
#include "md5.h"
#include "fileops.h"
#include "far_string.h"
#include "far_regex.h"
#include "list.h"
#include "utils.h"

ferr_t far_filedesc_archive(const char **dirlist, int ndirs,
				const char *regexp, int fd)
{
	buf_t	*retval	= (buf_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(dirlist && ndirs && (fd != -1))
	{
		result = _far_archive(dirlist, ndirs, regexp, &retval,
					_FAR_STREAM_FD, fd);
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	return result;
}

ferr_t far_buffer_archive(const char **dirlist, int ndirs, const char *regexp,
			void **buffer, uint64_t *bufsize)
{
	buf_t	*retval	= (buf_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(dirlist && ndirs && buffer && bufsize)
	{
		result = _far_archive(dirlist, ndirs, regexp, &retval,
					_FAR_STREAM_BUF, 0);

		if(_far_noerror(result))
		{
			(*buffer) = far_buffer(retval);
			(*bufsize) = far_bufsize(retval);
			free(retval);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t _far_archive(const char **dirlist, int ndirs, const char *regexp,
			buf_t **buffer, int buftype, int fd)
{
	int		count	= 0;
	regex_t		*preg	= (regex_t *)NULL;
	far_list_t	*head	= (far_list_t *)NULL;
	far_list_t	**ptr	= (far_list_t **)&head;
	char		*olddir	= (char *)NULL;
	uint64_t	offset	= 0;
	char_t		first	= TRUE;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(dirlist && ndirs && *dirlist && buffer)
	{
		olddir = getcwd(NULL, 1024);

		if(!olddir)
		{
			_far_set_errcode(result, errno);
		}

		if(regexp && *regexp && _far_noerror(result))
		{
			preg = far_regcomp(regexp);

			if(!preg)
			{
				_far_set_errcode(result, _FAR_ERR_REGCOMP);
			}
		}

		for(count = 0 ; (count < ndirs) && _far_noerror(result)  ;
			count++)
		{
			if(dirlist[count])
			{
				result = far_collectinfo(ptr, dirlist[count],
							preg, &first);

				if(_far_noerror(result))
				{
					ptr = &((*ptr)->next);
				}
			}
		}

		if(_far_noerror(result))
		{
			result = far_set_offset(head, &offset);

			if(_far_noerror(result))
			{
				result = far_create_archive(buffer, head,
								buftype, fd);
			}
		}

		if(olddir)
		{
			far_chdir(olddir);
		}

		if(preg)
		{
			regfree(preg);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	if(preg)
	{
		free(preg);
	}

	RETURN(result);
}

ferr_t far_collectinfo(far_list_t **list, const char *dname, regex_t *preg,
			char_t *first)
{
	struct	stat	*stbuf		= (struct stat *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(dname && list)
	{
		stbuf = far_stat(dname);

		result = far_add_entry(list, dname, stbuf);

		if(_far_noerror(result))
		{
			if(S_ISDIR(stbuf->st_mode))
			{
				result = far_readdir(list, dname, preg, first);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}
	RETURN(result);
}

ferr_t far_readdir(far_list_t **list, const char *dname, regex_t *preg,
			char_t *first)
{
	DIR		*dp	= (DIR *)NULL;
	char		*olddir	= (char *)NULL;
	struct	dirent	*tmp	= (struct dirent *)NULL;
	far_list_t	*ptr	= (far_list_t *)NULL;
	struct	stat	stbuf;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(list && dname)
	{
		ptr = (*list);
		if(!(olddir = getcwd(NULL, 1024)))
		{
			_far_set_errcode(result, errno);
		}

		if((far_chdir(dname) != -1) && _far_noerror(result))
		{
			dp = opendir(".");

			if(dp)
			{
				errno = 0;
				while((tmp = readdir(dp)) &&
					_far_noerror(result))
				{
					if(!(strcmp(tmp->d_name, ".")) ||
						!(strcmp(tmp->d_name, "..")))
					{
						continue;
					}

					if(preg && !far_match(tmp->d_name,
								preg))
					{
						continue;
					}

					lstat(tmp->d_name, &stbuf);

					if(S_ISDIR(stbuf.st_mode) || (*first))
					{
						far_list_t	**ptr2	= NULL;
						int		flag	= 0;
/*
 * flag: I used a little hack here to make things work correctly
 * I will change this code in the future to make it nicer
 */

						if(S_ISDIR(stbuf.st_mode))
						{
							(*first) = TRUE;
							ptr2 = &(ptr->next);
							flag = 1;
						}
						else
						{
							(*first) = FALSE;
							ptr2 = &(ptr->sub);
						}
						result = far_collectinfo(
								ptr2,
								tmp->d_name,
								preg, first);

						if(_far_noerror(result))
						{
							if(!flag)
							{
								(*first) =
									FALSE;
								farlink_s(ptr);
								ptr = ptr->sub;
							}
							else
							{
								(*first) =
									FALSE;
								farlink_f(ptr);
								ptr =ptr->next;
							}
						}
					}
					else
					{
						result = far_collectinfo(
								&(ptr->next),
								tmp->d_name,
								preg, first);

						if(_far_noerror(result))
						{
							ptr->next->prev = ptr;
							ptr->next->parent =
								ptr->parent;
							ptr = ptr->next;
						}
					}
				}

				if(_far_noerror(result))
				{
					_far_set_errcode(result, errno);
				}
				closedir(dp);
			}
			else
			{
				_far_set_errcode(result, errno);
			}

			far_chdir(olddir);
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_add_entry(far_list_t **info, const char *dname, struct stat *stbuf)
{
	far_list_t	*ptr	= (far_list_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info && dname && stbuf)
	{
		ptr = (*info) = calloc(1, sizeof(far_list_t));

		if(ptr)
		{
			if(strlen(dname) < 256)
			{
				ptr->d_path = strdup(dname);
				result = far_statdup(ptr, stbuf);

				if(_far_noerror(result) &&
					S_ISLNK(stbuf->st_mode))
				{
					result = far_resolvelink(ptr);
				}
			}
			else
			{
				_far_set_errcode(result, _FAR_ERR_NAMETOOLONG);
			}
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_set_offset(far_list_t *list, uint64_t *offset)
{
	far_list_t			*ptr	= (far_list_t *)NULL;
	register	uint32_t	psize	= 0;
	register	uint32_t	lpsize	= 0;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(list && offset)
	{
		for(ptr = list ; ptr && _far_noerror(result) ; ptr = ptr->next)
		{
			result = far_paths_size(ptr);

			if(_far_noerror(result))
			{
				ptr->d_offset = (*offset);
				psize = ptr->d_pathsize;
				lpsize = ptr->d_lpathsize;
				(*offset) += psize + lpsize;
			}

			if(ptr->sub && _far_noerror(result))
			{
				result = far_set_offset(ptr->sub, offset);
			}
			else if(S_ISREG(ptr->d_info->fst_mode))
			{
				if(_far_noerror(result))
				{
					(*offset) += ptr->d_info->fst_size;
				}
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_create_archive(buf_t **buffer, far_list_t *list,
			int buftype, int fd)
{
	uint64_t	headsize	= 0;
	uint64_t	contentsize	= 0;
	uint64_t	bufsize		= 0;
	buf_t		*bufptr		= (buf_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buffer && list)
	{
		headsize = far_headsize(list);

		if(headsize)
		{
			contentsize = far_contentsize(list);

			bufsize = (
				sizeof(uint64_t)	+
				headsize		+
				contentsize		+
				16			+
				sizeof(fver_t)
			);

			bufptr = (*buffer) = far_bufalloc(bufsize, buftype,
							fd);

			if(!bufptr)
			{
				_far_set_errcode(result, errno);
			}

			if(_far_noerror(result))
			{
				uint64_t	tmp	= 0;

				tmp = far_sw64bit(headsize);
				far_write(bufptr, &tmp, 8);
				result = far_put_header(list, bufptr);
			}

			if(_far_noerror(result))
			{
				result = far_put_content(list, bufptr);

				if(_far_noerror(result))
				{
					result = far_put_arcksum(bufptr);

					if(_far_noerror(result))
					{
						far_put_ver(bufptr);
					}
				}
			}
		}
		else
		{
			_far_set_errcode(result, _FAR_ERR_EMPTYHEAD);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_put_header(far_list_t *list, buf_t *buffer)
{
	far_list_t	*ptr	= list;
	far_head_t	*hp	= (far_head_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(list && buffer)
	{
		hp = calloc(1, sizeof(far_head_t));
		if(!hp)
		{
			_far_set_errcode(result, errno);
		}

		for( ; ptr && _far_noerror(result) ; ptr = ptr->next)
		{
			result = far_make_head(ptr, &hp);

			if(_far_noerror(result))
			{
				result = far_write(buffer, hp,
							sizeof(far_head_t));

				if(_far_noerror(result) && ptr->sub)
				{
					result = far_put_header(ptr->sub,
								buffer);
				}
			}
		}

		if(hp)
		{
			free(hp);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_put_content(far_list_t *list, buf_t *buffer)
{
	far_list_t	*ptr	= list;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(list && buffer)
	{
		for( ; ptr && _far_noerror(result) ; ptr = ptr->next)
		{
			if(S_ISREG(ptr->d_info->fst_mode))
			{
				result = far_put_file(ptr, buffer);
			}
			else
			{
				result = far_put_info(ptr, buffer);
			}

			if(_far_noerror(result))
			{
				if(ptr->sub)
				{
					result = far_put_content(ptr->sub,
								buffer);
				}
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_put_info(far_list_t *info, buf_t *buffer)
{
	char		*path	= (char *)NULL;
	char		*lpath	= (char *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info && buffer)
	{
		path = far_build_path(info);

		if(!path)
		{
			_far_set_errcode(result, errno);
		}

		if(_far_noerror(result))
		{
			result = far_write(buffer, path, strlen(path));

			if(_far_noerror(result))
			{
				if(info->d_lpath)
				{
					lpath = info->d_lpath;
					result = far_write(buffer, lpath,
								strlen(lpath));
					info->d_lpathsize = strlen(lpath);
				}
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_put_file(far_list_t *info, buf_t *buffer)
{
	char		*path	= (char *)NULL;
	char		*lpath	= (char *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info && buffer)
	{
		path = far_build_path(info);

		if(!path)
		{
			_far_set_errcode(result, errno);
		}

		if(_far_noerror(result))
		{
			result = far_write(buffer, path, strlen(path));

			if(_far_noerror(result))
			{
				info->d_pathsize = strlen(path);

				if(info->d_lpath)
				{
					lpath = info->d_lpath;
					result = far_write(buffer, lpath,
								strlen(lpath));
					info->d_lpathsize = strlen(lpath);
				}
			}
		}

		if(_far_noerror(result) && info->d_info->fst_size)
		{
			char_t	*d_cksum	= (char_t *)&(info->d_cksum[0]);
			result = far_append_file(path, info->d_info->fst_size,
					buffer, &(d_cksum));
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_put_arcksum(buf_t *buffer)
{
	char_t	*cksum	= (char_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buffer)
	{
		buffer->bufsize-=18; /* || an orrible hack */
		result = far_md5digest(buffer, &cksum);
		buffer->bufsize+=18; /* || */

		if(_far_noerror(result))
		{
			result = far_write(buffer, cksum, 16);
			free(cksum);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_make_head(far_list_t *info, far_head_t **hp)
{
	far_head_t	*p	= (far_head_t *)NULL;
	char_t		*hash	= (char_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info && hp)
	{
		p = (*hp) = calloc(1, sizeof(far_head_t));

		if(p)
		{
			far_statcpy(&(p->d_info), info->d_info);
			p->d_offset = far_sw64bit(info->d_offset);
			p->d_pathsize = far_sw32bit(info->d_pathsize);
			p->d_lpathsize = far_sw32bit(info->d_lpathsize);
			result = far_memcpy(p->d_cksum, info->d_cksum, 16);

			if(_far_noerror(result))
			{
				result = _far_make_namehash(info->d_path,
								&hash);

				if(_far_noerror(result))
				{
					result = far_memcpy(p->d_hash, hash,
								16);
				}
			}
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

void far_statcpy(far_stat_t *dest, far_stat_t *src)
{
	if(dest && src)
	{
		dest->fst_uid = far_sw32bit(src->fst_uid);
		dest->fst_gid = far_sw32bit(src->fst_gid);
		dest->fst_mode = far_sw32bit(src->fst_mode);
		dest->fst_atime = far_sw64bit(src->fst_atime);
		dest->fst_mtime = far_sw64bit(src->fst_mtime);
		dest->fst_rdev = far_sw64bit(src->fst_rdev);
		dest->fst_size = far_sw64bit(src->fst_size);
	}
}

ferr_t far_put_ver(buf_t *buffer)
{
	unsigned	short	a = htons(FAR_MAJORVERSION);

	RETURN(far_write(buffer, &(a), 2));
}
