#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include "far_types.h"
#include "far.h"
#include "far_err.h"
#include "streams.h"
#include "far_string.h"
#include "md5.h"
#include "fileops.h"
#include "extract.h"
#include "utils.h"
#include "bits.h"

ferr_t far_buffer_extract(char_t *buffer, uint64_t size, const char *entry)
{
	buf_t	*stream	= (buf_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buffer && size)
	{
		if(entry && !(*entry))
		{
			_far_set_errcode(result, _FAR_ERR_INVAL);
		}
		else
		{
			result = far_init_bufstream(&stream, buffer, size);

			if(_far_noerror(result))
			{
				result = _far_extract(stream, entry);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_filedesc_extract(int fd, const char *entry)
{
	buf_t	*stream	= (buf_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(fd)
	{
		result = far_init_fdstream(&stream, fd);

		if(_far_noerror(result))
		{
			result = _far_extract(stream, entry);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_sockdesc_extract(int fd)
{
	buf_t	*stream	= (buf_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(fd)
	{
		result = far_init_sdstream(&stream, fd);

		if(_far_noerror(result))
		{
			result = _far_extract(stream, NULL);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t _far_extract(buf_t *buf, const char *entry)
{
	uint64_t	headsize	= 0;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf)
	{
		if(_far_isbuffer(buf))
		{
			result = far_md5check_buf(buf);
		}
		else if(_far_isfiledesc(buf))
		{
			result = far_md5check_fd(buf);
		}

		if(_far_noerror(result))
		{
			result = far_get_headsize(buf, &headsize);

			if(_far_noerror(result))
			{
				if(entry && *entry)
				{
					result = far_extract_single(buf, entry,
								headsize);
				}
				else
				{
					result = far_start_extract(buf,
								headsize);
				}
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

#define	far_continue_extract far_start_extract
ferr_t far_start_extract(buf_t *buf, uint64_t headsize)
{
	far_info_t	*info	= (far_info_t *)NULL;
	char		*opwd	= (char *)NULL;
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf && headsize)
	{
		result = far_get_finfo(buf, &info, headsize);

		if(_far_noerror(result))
		{
			result = far_extract_entry(buf, info);

			if(_far_noerror(result) &&
			(far_bufindex(buf) < (headsize+sizeof(headsize))))
			{
				if(S_ISDIR(info->d_info->fst_mode))
				{
					opwd = getcwd(NULL, 1024);
//					chdir(info->d_path);
				}
				far_free_info(info);
				result = far_continue_extract(buf, headsize);
				if(S_ISDIR(info->d_info->fst_mode))
				{
//					chdir(opwd);
				}
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_extract_single(buf_t *buf, const char *entry, uint64_t headsize)
{
	far_info_t	*info	= (far_info_t *)NULL;
	char		*opwd	= (char *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf && entry && *entry && headsize)
	{
		result = far_find_entry(buf, entry, &info);

		if(_far_noerror(result))
		{
			result = far_extract_entry(buf, info);

			if(_far_noerror(result) &&
				S_ISDIR(info->d_info->fst_mode))
			{
				opwd = getcwd(NULL, 1024);
//				chdir(info->d_path);
				far_free_info(info);

				entry = far_append_str((char *)entry, "/");
				if(entry)
				{
					result = far_extract_sim(buf, entry,
								headsize);
				}
//				chdir(opwd);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_extract_sim(buf_t *buf, const char *entry, uint64_t headsize)
{
	far_info_t	*info	= (far_info_t *)NULL;
	char		*opwd	= (char *)NULL;
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf && entry && *entry && headsize)
	{
		result = far_find_simentry(buf, entry, &info, headsize);

		if(_far_noerror(result))
		{
			result = far_extract_entry(buf, info);

			if(_far_noerror(result) &&
				S_ISDIR(info->d_info->fst_mode))
			{
				opwd = getcwd(NULL, 1024);
//				chdir(info->d_path);
				far_free_info(info);

				if(entry)
				{
					result = far_extract_sim(buf, entry,
								headsize);
				}
//				chdir(opwd);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_find_entry(buf_t *buf, const char *d_name, far_info_t **info)
{
	char_t		*hash	= (char_t *)NULL;
	far_head_t	head;
	far_head_t	*ptr	= (far_head_t *)&head;
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info && buf && d_name && *d_name)
	{
		result = _far_make_namehash(d_name, &hash);

		if(_far_noerror(result))
		{
			(*info) = NULL;
			while(!(*info) && _far_noerror(result))
			{
				result = far_get_head(buf, &ptr);

				if(_far_noerror(result) && !memcmp(hash,
					head.d_hash, 16))
				{
					result = far_head2info(&head, (*info));

					if(_far_noerror(result))
					{
						result = far_getnames(buf,
								(*info));
					}
				}
			}
			_SFREE(hash);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_find_simentry(buf_t *buf, const char *d_name, far_info_t **info,
			uint64_t headsize)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info && buf && d_name && *d_name)
	{
		(*info) = NULL;
		while(!(*info) && _far_noerror(result))
		{
			result = far_get_finfo(buf, info, headsize);

			if(_far_noerror(result))
			{
				if(strncmp(d_name, (*info)->d_path,
						strlen(d_name)))
				{
					far_free_info((*info));
					(*info) = NULL;
				}
			}
			else
			{
				far_free_info((*info));
				(*info) = NULL;
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_extract_entry(buf_t *buf, far_info_t *info)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf && info)
	{
		if(S_ISDIR(info->d_info->fst_mode))
		{
			result = far_mkdir(info);
		}
		else if(S_ISLNK(info->d_info->fst_mode))
		{
			result = far_symlink(info);
		}
		else if(S_ISCHR(info->d_info->fst_mode) ||
			S_ISBLK(info->d_info->fst_mode) ||
			S_ISFIFO(info->d_info->fst_mode))
		{
			result = far_mknod(info);
		}
		else if(S_ISREG(info->d_info->fst_mode))
		{
			result = far_create_file(buf, info);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_get_finfo(buf_t *buf, far_info_t **info, uint64_t headsize)
{
	far_head_t	*head	= (far_head_t *)NULL;
	far_info_t	*ptr	= (far_info_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf && info)
	{
		ptr = (*info) = calloc(1, sizeof(far_info_t));

		if(ptr)
		{
			result = far_get_head(buf, &head);

			if(_far_noerror(result))
			{
				result = far_head2info(head, ptr);

				if(_far_noerror(result))
				{
					ptr->d_offset += headsize +
							sizeof(headsize);
					result = far_getnames(buf, ptr);
				}
			}

			_SFREE(head);
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_get_head(buf_t *buf, far_head_t **head)
{
	far_head_t	*ptr	= (far_head_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf && head)
	{
		ptr = (*head) = calloc(1, sizeof(far_head_t));

		if(ptr)
		{
			result = far_read(buf, ptr, sizeof(far_head_t), NULL);

			if(_far_noerror(result))
			{
				result = far_bconv_head(ptr);
			}
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_head2info(far_head_t *head, far_info_t *info)
{
	char_t	*hash	= (char_t *)NULL;
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(head && info)
	{
		info->d_info = calloc(1, sizeof(far_stat_t));

		if(info->d_info)
		{
			result = far_memcpy(info->d_info, &(head->d_info),
						sizeof(far_stat_t));

			if(_far_noerror(result))
			{
				info->d_pathsize = head->d_pathsize;
				info->d_lpathsize = head->d_lpathsize;
				info->d_offset = head->d_offset;

				result = far_memcpy(info->d_cksum,
						head->d_cksum, 16);

				if(_far_noerror(result))
				{
//					result = _far_make_namehash(info->d_path, &hash);
					result = far_memcpy(info->d_hash,
							head->d_hash, 16);

					if(_far_noerror(result))
					{
//						result = far_memcpy(head->d_hash, hash, 16);
					}
				}
			}
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_getnames(buf_t *buf, far_info_t *info)
{
	uint64_t	pos	= 0;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info)
	{
		info->d_path = calloc(1, (info->d_pathsize + 1));

		if(info->d_path)
		{
			if(info->d_lpathsize)
			{
				info->d_lpath = calloc(1,
						(info->d_lpathsize + 1));
			}

			if(info->d_lpath || !(info->d_lpathsize))
			{
				pos = _far_bufpos(buf);
				result = _far_bufseek(buf, info->d_offset);

				if(_far_noerror(result))
				{
					result = far_read(buf, info->d_path,
						info->d_pathsize, NULL);
				}

				if(_far_noerror(result) && info->d_lpathsize)
				{
					result = far_read(buf, info->d_lpath,
						info->d_lpathsize, NULL);
				}

				_far_bufseek(buf, pos);
			}
			else
			{
				_far_set_errcode(result, errno);
			}
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_mkdir(far_info_t *info)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info && info->d_path && *(info->d_path))
	{
		result = far_make_path(info->d_path, info->d_info->fst_mode,
				info->d_info->fst_uid, info->d_info->fst_gid);
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_mknod(far_info_t *info)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info && info->d_path && *(info->d_path))
	{
		result = far_makedev(info->d_path, info->d_info->fst_rdev);

		if(_far_noerror(result))
		{
			result = far_chown(info);

			if(_far_noerror(result))
			{
				result = far_chmod(info);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_symlink(far_info_t *info)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info && info->d_lpath && *(info->d_lpath))
	{
		if(!symlink(info->d_lpath, info->d_path))
		{
			result = far_chown(info);

			if(_far_noerror(result))
			{
				result = far_chmod(info);
			}
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_create_file(buf_t *buf, far_info_t *info)
{
	uint64_t	pos		= 0;
	uint64_t	bytes		= 0;
	uint64_t	last		= 0;
	uint64_t	size		= _FAR_FBUF_SIZE;
	int		fd		= 0;
	int		cantopen	= 0;
	char_t		*buffer		= (char_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf && info && info->d_path && *(info->d_path))
	{
		pos = _far_bufpos(buf);

		size = ((size > info->d_info->fst_size) ?
				info->d_info->fst_size : size);

		buffer = calloc(1, _FAR_FBUF_SIZE);

		if(!buffer)
		{
			_far_set_errcode(result, errno);
		}

		if(_far_noerror(result))
		{
			_far_bufseek(buf, info->d_offset + info->d_pathsize +
					info->d_lpathsize);
			result = far_md5check_file(buf, info->d_cksum);
		}

		if(_far_noerror(result))
		{
			fd = open(info->d_path, O_WRONLY|O_TRUNC|O_CREAT);

			if(fd == -1)
			{
				cantopen = 1;
				_far_set_errcode(result, errno);
			}
		}

		buf->bufsize -= 18;	/* that orrible hack again */
		while(_far_noerror(result))
		{
			result = far_read(buf, buffer, size, &last);
			bytes += last;

			if(_far_noerror(result))
			{
				result = far_fwrite(fd, buffer, last);

				if(_far_noerror(result) &&
					(bytes >= info->d_info->fst_size))
				{
					_far_set_errcode(result, _FAR_ERR_EOF);
				}
			}
		}
		buf->bufsize+=18;

		if(_far_err_comp(result, _FAR_ERR_EOF))
		{
			_far_err_clear(result);
		}

		if(!cantopen)
		{
			close(fd);
			errno = 0;	/* we don't care about errno here */
		}

		_far_bufseek(buf, pos);

		if(_far_noerror(result))
		{
			result = far_chown(info);

			if(_far_noerror(result))
			{
				result = far_chmod(info);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_get_headsize(buf_t *buf, uint64_t *hsize)
{
	uint64_t	pos	= 0;
	uint64_t	bytes	= 0;
	buf_t		*tmp	= (buf_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf && hsize)
	{
		pos = _far_bufpos(buf);

		if(pos >= 0)
		{
			result = _far_bufseek(buf, 0L);

			if(_far_noerror(result))
			{
				tmp = far_bufalloc(sizeof(uint64_t),
						_FAR_STREAM_BUF, 0);

				if(tmp)
				{
					result = far_read(buf, hsize,
						sizeof(bytes), &bytes);

					if(_far_noerror(result) &&
						(bytes != sizeof(bytes)))
					{
						_far_set_errcode(result,
							_FAR_ERR_BADARCH);
					}
				}
				else
				{
					_far_set_errcode(result, errno);
				}
			}
//			_far_bufseek(buf, pos);
		}
		else
		{
			_far_set_errcode(result, _FAR_ERR_SEGFLT);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_md5check_file(buf_t *buf, char_t *cksum)
{
	uint64_t	pos		= 0;
	uint64_t	size		= 0;
	char_t		*buffer		= (char_t *)NULL;
	char_t		*c_cksum	= (char_t *)NULL;
	md5_state_t	*md5		= (md5_state_t *)NULL;
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf && cksum)
	{
		pos = _far_bufpos(buf);

		buffer = calloc(1, _FAR_FBUF_SIZE);

		if(!buffer)
		{
			_far_set_errcode(result, errno);
		}
		else
		{
			md5 = calloc(1, sizeof(md5_state_t));

			if(!md5)
			{
				_far_set_errcode(result, errno);
			}
			else
			{
				c_cksum = calloc(1, (16 * sizeof(char_t)));

				if(!c_cksum)
				{
					_far_set_errcode(result, errno);
				}
			}
		}

		while(_far_noerror(result))
		{
			result = far_read(buf, buffer, _FAR_FBUF_SIZE, &size);

			if(_far_noerror(result))
			{
				md5_append(md5, buffer, size);
			}

			if(_far_noerror(result) && (size != _FAR_FBUF_SIZE))
			{
				_far_set_errcode(result, _FAR_ERR_EOF);
			}
		}

		if(_far_err_comp(result, _FAR_ERR_EOF))
		{
			_far_err_clear(result);
		}

		if(_far_noerror(result))
		{
			md5_finish(md5, c_cksum);

			if(memcmp(c_cksum, cksum, 16))
			{
//				_far_set_errcode(result, _FAR_ERR_BADFILECRC);
			}
		}
		_far_bufseek(buf, pos);
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}
