#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "far_types.h"
#include "far.h"
#include "md5.h"
#include "streams.h"
#include "fileops.h"

ferr_t far_statdup(far_list_t *info, struct stat *stbuf)
{
	far_stat_t	*ptr	= (far_stat_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info && stbuf)
	{
		info->d_info = calloc(1, sizeof(far_stat_t));

		if(info->d_info)
		{
			ptr = info->d_info;
			ptr->fst_uid = stbuf->st_uid;
			ptr->fst_uid = stbuf->st_gid;
			ptr->fst_mode = stbuf->st_mode;
			ptr->fst_atime = stbuf->st_atime;
			ptr->fst_mtime = stbuf->st_mtime;
			ptr->fst_rdev = stbuf->st_rdev;
			ptr->fst_size = stbuf->st_size;
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_resolvelink(far_list_t *info)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info)
	{
		if((info->d_lpath = calloc(1, 1032)))
		{
			if(!readlink(info->d_path, info->d_lpath, 1024))
			{
				info->d_lpath = realloc(info->d_lpath,
						(strlen(info->d_lpath) + 1));
			}
			else
			{
				_far_set_errcode(result, errno);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_append_file(char *path, uint64_t size, buf_t *buffer,
			char_t **cksum)
{
	int	fd	= 0;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(path && buffer)
	{
		fd = open(path, O_RDONLY);

		if(fd != -1)
		{
			result = far_fread(fd, buffer, size, NULL);

			if(_far_noerror(result))
			{
				result = far_md5digest(buffer, cksum);
			}
			close(fd);
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_md5digest(buf_t *buffer, char_t **digest)
{
	md5_state_t	*md5	= (md5_state_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buffer && digest)
	{
		md5 = calloc(1, sizeof(md5_state_t));

		if(md5)
		{
			(*digest) = calloc(1, 16);

			if(*digest)
			{
				far_md5_append(md5, buffer);
				md5_finish(md5, (*digest));
				free(md5);
			}
			else
			{
				_far_set_errcode(result, errno);
			}
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

struct  stat *far_stat(const char *path)
{
	struct	stat	*result	= (struct stat *)NULL;

	if(path)
	{
		result = calloc(1, sizeof(struct stat));

		if(result)
		{
			if(lstat(path, result) == -1)
			{
				free(result);
				result = NULL;
			}
		}
	}

	RETURN(result);
}

ferr_t far_md5_append(md5_state_t *md5, buf_t *buffer)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(md5 && buffer)
	{
		md5_append(md5, far_buffer(buffer), far_bufsize(buffer));
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_fwrite(int fd, void *buf, uint64_t size)
{
	register	uint64_t	bytes	= 0;
	register	uint64_t	last	= 0;
	register	uint64_t	tow	= _FAR_FBUF_SIZE;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buf)
	{
		if(size)
		{
			tow = ((tow > size) ? size : tow);
			while((bytes < size) && (last = write(fd, buf,
							tow)))
			{
				bytes +=last;
			}

			if(bytes != size)
			{
				_far_set_errcode(result, errno);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_fread(int fd, buf_t *buffer, uint64_t size, uint64_t *retsize)
{
	register	uint64_t	bytes	= 0;
	register	uint64_t	last	= 0;
	char_t				*buf	= (char_t *)NULL;

	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(buffer)
	{
		if(size)
		{
			buf = calloc(1, _FAR_FBUF_SIZE);

			if(buf)
			{
				while((bytes < size) && _far_noerror(result) &&
					(last = read(fd, buf, _FAR_FBUF_SIZE)))
				{
					if(last != -1)
					{
						result = far_write(buffer,
								buf, last);
						if(_far_noerror(result))
						{
							bytes +=last;
						}
					}
					else
					{
						_far_set_errcode(result,errno);
					}
				}

				if(!last)
				{
					_far_set_errcode(result, _FAR_ERR_EOF);
				}
				else if(retsize)
				{
					(*retsize) = bytes;
				}
			}
			else
			{
				_far_set_errcode(result, errno);
			}

			if(_far_noerror(result) && (bytes != size))
			{
				_far_set_errcode(result, errno);
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_make_path(char *dir, mode_t perm, int uid, int gid)
{
	struct	stat		fbuf;
	int			ret		= 0,
				i		= 0;
	char			path[256]	= { 0 };
	char_t			finished	= FALSE;
	unsigned	int	len		= 0;
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	errno = 0;
	
	if((len=strlen(dir)) == 0 || !dir)
		RETURN(-1);
	
	strcpy(path, dir);
	if(path[len-1] == '/')
	{
		len--;
		path[len] = '\0';
	}
	
	/* Find the first directory that exists */
	
	i = len;
	while(!finished && _far_noerror(result))
	{
		if(((ret = stat(path, &fbuf)) == -1) &&
			(errno == ENOENT))
		{
			errno = 0;
			while((i > 0) && (path[i] != '/') && (path[i] != '\0'))
				i--;
			path[i] = '\0';
			if(!i)
			{
				finished = TRUE;
			}
			else    /* Directory exists */
			{
				finished = TRUE;
			}
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	
	/* At this point we have the first existing directory */
	
		if(i != len)   /* ALL of directory path already exists */
		{
			while((i != len) && _far_noerror(result))
			{
		/* Move to next dir */
	
				i = strlen(path);
				if(i != len)
				{
					path[i] = dir[i];
					i++;
					if((mkdir(path, perm) == -1) &&
						(errno != EEXIST))
					{
						_far_set_errcode(result, errno);
					}
					else
					{
						errno=0;
						if(chown(path, uid, gid))
						{
							_far_set_errcode(result, errno);
						}
					}
				}
			}
		}
		else
		{
			finished = TRUE;
		}
	}

	RETURN(result);
}

ferr_t far_chown(far_info_t *info)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info)
	{
		if(chown(info->d_path, info->d_info->fst_uid,
			info->d_info->fst_gid) == -1)
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_chmod(far_info_t *info)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info)
	{
		if(chmod(info->d_path, info->d_info->fst_mode) == -1)
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_lseek(int fd, uint64_t offset, int whence)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if((fd >= 0) && (offset >= 0))
	{
		if(lseek(fd, offset, whence) == -1)
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_makedev(char *path, uint64_t rdev)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(path && *path)
	{
		if(mknod((const char *)path, 0, rdev) == -1)
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}
