#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "far_types.h"
#include "far.h"
#include "far_err.h"
#include "far_string.h"
#include "utils.h"
#include "md5.h"

char *far_append_str(char *str1, const char *str2)
{
	char	*result	= (char *)NULL;

	if(str1 && str2 && *str2)
	{
		str1 = realloc((void *)str1, (strlen((const char *)str1) + strlen(str2) + 1));

		if(str1)
		{
			result = strcat((char *)str1, (const char *)str2);
		}
	}

	RETURN(result);
}
uint64_t far_num_nodes(far_list_t *list)
{
	uint64_t	result	= 0;
	far_list_t	*ptr	= (far_list_t *)NULL;

	if(list)
	{
		for(ptr = list ; ptr ; ptr = ptr->next)
		{
			if(ptr->sub)
			{
				result += far_num_nodes(ptr->sub);
			}
			result++;
		}
	}

	RETURN(result);
}

uint64_t far_contentsize(far_list_t *list)
{
	uint64_t	result	= 0;
	far_list_t	*ptr	= (far_list_t *)NULL;

	if(list)
	{
		for(ptr = list ; ptr ; ptr = ptr->next)
		{
			result += ptr->d_pathsize + ptr->d_lpathsize;
			if(ptr->sub)
			{
				result += far_contentsize(ptr->sub);
			}
			if(S_ISREG(ptr->d_info->fst_mode))
			{
				result += ptr->d_info->fst_size;
			}
		}
	}

	RETURN(result);
}

ferr_t far_paths_size(far_list_t *info)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info && info->d_path)
	{
		info->d_pathsize = strlen(far_build_path(info));
		if(info->d_lpath)
		{
			info->d_lpathsize = strlen(info->d_lpath);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_free_info(far_info_t *info)
{
	far_info_t	*ptr	= (far_info_t *)NULL;
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(info)
	{
		_SFREE(info->d_info);
		_SFREE(info->d_path);
		_SFREE(info->d_lpath);

		if(info->sub)
		{
			result = far_free_info(info->sub);

			if(_far_noerror(result))
			{
				ptr = info->next;
				_SFREE(info);
				if(ptr)
				{
					result = far_free_info(ptr);
				}
			}
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t _far_make_namehash(const char *d_name, char_t **hash)
{
	md5_state_t	*md5	= (md5_state_t *)NULL;
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(d_name && hash)
	{
		(*hash) = calloc(1, (16 * sizeof(char_t)));

		if((*hash))
		{
			md5 = calloc(1, sizeof(md5_state_t));

			if(md5)
			{
				md5_append(md5, (char_t *)d_name, strlen(d_name));
				md5_finish(md5, (*hash));
				_SFREE(md5);
			}
			else
			{
				_SFREE(*hash);
				_far_set_errcode(result, errno);
			}
		}
		else
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

ferr_t far_memcpy(void *dest, void *src, uint64_t len)
{
	_far_returnvalue(result, _FAR_ERR_SUCCESS);

	if(dest && src && (len > 0))
	{
		if(!memcpy(dest, src, len))
		{
			_far_set_errcode(result, errno);
		}
	}
	else
	{
		_far_set_errcode(result, _FAR_ERR_INVAL);
	}

	RETURN(result);
}

int far_chdir(const char *p)
{
	printf("Entering: %s\n", p);
	RETURN(chdir(p));
}

